"""
    This file exposes the main API for this library, the microscope Manager
"""
# Local imports
from c4v.scraper.persistency_manager.base_persistency_manager import (
    BasePersistencyManager,
)
from c4v.scraper.persistency_manager.sqlite_storage_manager import SqliteManager
from c4v.scraper.scraped_data_classes.scraped_data import ScrapedData
from c4v.scraper.scraper import bulk_scrape, _get_scraper_from_url, scrape
from c4v.scraper.settings import INSTALLED_CRAWLERS, SUPPORTED_DOMAINS
from c4v.config import settings
from c4v.microscope.metadata import Metadata

# Python imports
import os
from typing import Dict, List, Iterable, Callable, Tuple, Any, Union
from pathlib import Path
import sys

# Configs
from c4v.config import settings

DEFAULT_DB = settings.local_sqlite_db or os.path.join(
    settings.c4v_folder, settings.local_sqlite_db_name
)

class Manager:
    """
        This object encapsulates shared behavior between our multiple components,
        allowing easy access to common operations. Note that this operations are intended to work 
        with a local db, so most of them will save data to some kind of storage. If no local db is 
        provided, some operations may not work
    """

    def __init__(
        self,
        persistency_manager: BasePersistencyManager,
        metadata: Metadata,
        local_files_path: str = settings.c4v_folder,
    ):
        self._persistency_manager = persistency_manager
        self._metadata = metadata
        self._local_files_path = local_files_path

    @property
    def local_files_path(self) -> str:
        """
            Path where files (.c4v folder) will be saved
        """
        return self._local_files_path

    @local_files_path.setter
    def local_files_path(self, new_path: str):
        self._local_files_path = new_path

    @property
    def persistency_manager(self) -> BasePersistencyManager:
        return self._persistency_manager

    def scrape(self, urls : Union[List[str], str] ) -> Union[List[ScrapedData], ScrapedData]:
        """
            Scrape the given url or set of urls and return its results
        """
        if not urls: # if nothing to process, return None
            return None 
        elif isinstance(urls, list) and urls and isinstance(urls[0], str):
            return bulk_scrape(urls)
        elif isinstance(urls, str):
            return scrape(urls)
        
        raise TypeError(f"Expected argument should be a list of urls as strings or a single string url. Given {type(urls)}")

    def crawl_and_scrape_for(self, crawler_names : List[str] = None, limit : int = -1, save_to_db = True) -> List[ScrapedData]:
        """
            Crawl and scrape data for the given list of crawlers
            Parameters:
                crawler_names : [str] = List of crawlers to use, defaults to all crawlers
                limit : int = Maximum ammount of urls to scrape, set to negative number for no limit (no recommended)
            Return:
                List of scraped data    
        """
        urls = self.crawl_new_urls_for(crawler_names=crawler_names, limit=limit, save_to_db=save_to_db)
        data = self.get_bulk_data_for(urls, save_to_db=save_to_db)
        return data

    def get_bulk_data_for(
        self, urls: List[str], should_scrape: bool = True, save_to_db = True
    ) -> List[ScrapedData]:
        """
            Retrieve scraped data for given url set if scrapable
            Parameters:
                urls : [str] = urls whose data is to be retrieved. If not available yet, then scrape it if requested so
                should_scrape : bool = if should scrape non-existent urls
                save_to_db : bool = if should save to db
            Return:
                List of gathered data, 
        """
        # just a shortcut
        db = self._persistency_manager

        # Separate scraped urls from non scraped
        not_scraped = db.filter_scraped_urls(urls) if db else urls
        
        # Scrape missing instances if necessary
        items = []
        if should_scrape and not_scraped:
            items = bulk_scrape(not_scraped)
            if save_to_db and db:
                db.save(items)

        urls = set(urls)
        return [sd for sd in db.get_all() if sd.url in urls] if db else items

    def get_data_for(self, url: str, should_scrape: bool = True) -> ScrapedData:
        """
            Get data for this url if stored or scrapable. May return none if could not
            find data for this url
            Parameters:
                url : str = url to be scraped
                should_scrape : bool = if should scrape this url if not available in db
        """
        data = self.get_bulk_data_for([url], should_scrape)
        return data[0] if data else None

    def scrape_pending(self, limit: int = -1):
        """
            Update DB by scraping rows with no scraped data, just the url
            Parameters:
                limit : int = how much measurements to scrape, set a negative number for no limit
        """
        if not self._persistency_manager:
            raise AttributeError("Storage scheme not configured for microscope Manager instance, provide a db scheme to use this function")

        db = self._persistency_manager

        scrape_urls = [d.url for d in db.get_all(limit=limit, scraped=False)]

        scraped = bulk_scrape(scrape_urls)

        db.save(scraped)

    def get_all(self, limit: int = -1, scraped: bool = None) -> Iterable[ScrapedData]:
        """
            Return all ScrapedData instances available, up to "limit" rows. If scraped = true, then
            return only scraped instances. If scraped = false, return only non scraped. Otherwise, return anything
            Parameters:
                limit : int = max ammount of rows to return. If negative, there's no limit of rows
                scraped : bool = If instances should be scraped or not. If true, all of them should be scraped. If false, none of 
                                them should be scraped. If None, there's no restriction
            Return:
                An iterator returning available rows
        """
        if not self._persistency_manager:
            raise AttributeError("Storage scheme not configured for microscope Manager instance, provide a db scheme to use this function")

        return self._persistency_manager.get_all(limit, scraped)

    def crawl_and_process_new_urls_for(
        self,
        crawler_names: List[str] = None,
        post_process: Callable[[List[str]], None] = None,
        limit=-1,
        save_to_db = True
    ):
        """
            Crawl for new urls using the given crawlers only
            Parameters:
                crawler_names : [str] = names of crawlers to be ran when this function is called. If no list is provided, then 
                                        all crawlers will be used
                post_process : ([str]) -> None = Function to call over new elements as they come
                limit        : int = Max amount of urls to save
                save_to_db : bool = If should save to local
        """
        db = self._persistency_manager
        limit = limit if limit >= 0 else sys.maxsize

        # set up limit if necessary
        limit = limit if limit >= 0 else sys.maxsize

        class Counter:
            def __init__(self):
                self.count = 0

            def add(self, x: int):
                self.count += x

        counter = Counter()

        # Function to process urls as they come
        def save_urls(urls: List[str]):

            urls = urls[:limit - counter.count]

            if db:
                urls = db.filter_known_urls(urls)
                datas = [ScrapedData(url=url) for url in urls]

                if save_to_db: db.save(datas)

            # Update how much elements have beed added so far
            counter.add(len(urls))

            # Call any callback function
            if post_process:
                post_process(urls)

        # Function to tell if the crawling process should stop
        def should_stop() -> bool:
            return counter.count >= limit

        # if no list provided, default to every crawler
        if crawler_names == None:
            crawler_names = [c.name for c in INSTALLED_CRAWLERS]

        # Instantiate crawlers to use
        crawlers_to_run = [
            crawler() for crawler in INSTALLED_CRAWLERS if crawler.name in crawler_names
        ]

        # Warns the user if no crawler could be found
        if not crawlers_to_run:
            given_crawler_list = "\n".join([f"\t* {c}" for c in crawler_names] or ["\t<No crawlers>"])
            actual_crawler_list = "\n".join([f"\t* {c.name}" for c in INSTALLED_CRAWLERS] or ["\t<No crawlers>"])
            print(f"[Warning] No crawler available to run. Given crawler list: \n{given_crawler_list}\n Available crawlers: \n{actual_crawler_list}", file=sys.stderr)

        # crawl for every crawler
        for crawler in crawlers_to_run:
            crawler.crawl_and_process_urls(save_urls, should_stop)

    def crawl_new_urls_for( self,
        crawler_names: List[str] = None,
        limit=-1,
        save_to_db : bool = True
    ):
        """
            Crawl new urls for the given list of crawlers:
            Parameters:
                crawler_names : [str] = list of names of scrapers to use when crawling
        """
        items = []
        
        def collect_urls(urls : List[str]):
            items.extend(urls)

        self.crawl_and_process_new_urls_for(crawler_names=crawler_names, limit=limit, post_process=collect_urls, save_to_db=save_to_db)

        return items

    def split_non_scrapable(self, urls: List[str]) -> Tuple[List[str], List[str]]:
        """
            splits url list in two list, one having a list of scrapable urls, 
            and another one with only non-scrapable
            Parameters:
                urls : [str] = List of urls to be split
            Return:
                A tuple with two lists, the first one with only scrapable urls from input list,
                the second one with non-scrapable ones
        """
        scrapable, non_scrapable = [], []

        for url in urls:
            try:
                _get_scraper_from_url(url)
                scrapable.append(url)
            except ValueError:
                non_scrapable.append(url)

        return scrapable, non_scrapable

    @classmethod
    def from_local_sqlite_db(cls, db_path: str, metadata: Union[str, Metadata]):
        """
            Create a new instance using an SQLite local db
            metadata : str | Metadata = Path to file with metadata or metadata instance itself
        """
        if isinstance(metadata, Metadata):
            pass  # Everything ok, just keep going
        elif isinstance(metadata, str):
            # If string is provided, interpret it as a filepath
            metadata = Metadata.from_json(metadata)
        else:
            raise TypeError(
                f"metadata field should be path to metadata files or metadata instance. Provided object's type: {type(metadata)}"
            )

        db = SqliteManager(db_path)
        return cls(db, metadata)

    @classmethod
    def from_default(
        cls, db_path: str = None, metadata: str = None, local_files_path: str = None, db : BasePersistencyManager = None
    ):
        """
            Create a Manager instance using files from the default `C4V_FOLDER`.
            Parameters:
                - db_path : str = Path to a sqlite db file to use as database.
                - metadata : str = Path to a valid metadata json file 
                - local_files_path : str = path to the .c4v folder to use for local files
                - db : BasePersistencyManager = persistency manager object to use as db manager. When provided, it overrides the db_path 
                                                parameter

        """

        # Set up db
        db = db or SqliteManager(
                    db_path
                    or str(
                        Path(
                            local_files_path or settings.c4v_folder,
                            settings.local_sqlite_db_name,
                        )
                    )
                )

        # Set up metadata
        metadata = Metadata.from_json(metadata) if metadata else Metadata()

        return cls(db, metadata, local_files_path or settings.c4v_folder)

    def run_classification_from_experiment(
        self, branch: str, experiment: str, data: List[ScrapedData]
    ) -> List[Dict[str, Any]]:
        """
            Classify given data instance list, returning its metrics
            Parameters:
                branch : str = branch name of model to use
                experiment : str = experiment name storing model
                data : [ScrapedData] = Instance to be classified
            Return:
                A List of dicts with the resulting scraped data correctly labelled
                and its corresponding scores tensor for each possible label. Available fields:
                    + data : ScrapedData = resulting data instance after classification
                    + scores : torch.Tensor = Scores for each label returned by the classifier
        """
        from c4v.classifier.classifier_experiment import ClassifierExperiment

        classifier_experiment = ClassifierExperiment.from_branch_and_experiment(
            branch, experiment
        )

        return classifier_experiment.classify(data)

    def run_pending_classification_from_experiment(
        self, branch: str, experiment: str, save: bool = True, limit: int = -1
    ) -> List[Dict[str, Any]]:
        """
            Classify data pending for classification in local db, returning the obtained results and saving it 
            to database. 
            Parameters:
                branch : str = branch name
                experiment : str = experiment name
                save : bool = if should store results in db
                limit : maximum number of rows to classify
            Return:
                A List of dicts with the resulting scraped data correctly labelled
                and its corresponding scores tensor for each possible label. Available fields:
                    + data : ScrapedData = resulting data instance after classification
                    + scores : torch.Tensor = Scores for each label returned by the classifier
        """
        # Parse limit
        limit = limit if limit >= 0 else sys.maxsize

        # Request at the most "limit" instances
        data = list(
            x for x in self.persistency_manager.get_all(scraped=True) if not x.label
        )[:limit]

        # classify
        results = self.run_classification_from_experiment(branch, experiment, data)

        # Save if requested so
        if save:
            self.persistency_manager.save((x["data"] for x in results))

        return results

    def explain_for_experiment(
        self,
        branch: str,
        experiment: str,
        sentence: str,
        html_file: str = None,
        additional_label: str = None,
    ) -> Dict[str, Any]:
        """
            Explain a sentence using the given branch and experiment
            Parameters:
                branch : str = branch name of model to use
                experiment : str = experiment name storing model
                sentence : str = sentence to explain
                additional_label : str = Label to include in expalantion. If the predicted 
                                         label is different from this one, then explain how 
                                         much this label was contributing to its corresponding value, 
                                         ignored if not provided
            Return:
                Dict with explaination data
        """

        from c4v.classifier.classifier_experiment import ClassifierExperiment

        classifier_experiment = ClassifierExperiment.from_branch_and_experiment(
            branch, experiment
        )

        return classifier_experiment.explain(
            sentence, html_file, additional_label=additional_label
        )

    def get_classifier_labels(self) -> List[str]:
        """
            Get list of possible labels for classifier
            Return:
                List with possible output labels for the classifier
        """
        from c4v.classifier.classifier import Classifier
        return Classifier.get_labels()

    @staticmethod
    def get_available_crawlers() -> List[str]:
        """
            List of usable names of crawlers
        """
        return [c.name for c in INSTALLED_CRAWLERS]

    @classmethod
    def scrapable_domains() -> List[str]:
        """
            List of scrapable domains, it may not match with the crawlable sites
        """
        return SUPPORTED_DOMAINS

    def should_retrain_base_lang_model(
        self,
        lang_model, #: LanguageModel,
        db_manager: BasePersistencyManager = None,
        eval_dataset_size: int = 250,
        min_loss: float = settings.default_lang_model_min_loss,
        should_retrain_fn: Callable[[float], bool] = None,
        fields: List[str] = ["title"],
    ) -> bool:
        """
            If should retrain a base language model based on its loss on the given dataset. If a persistency manager is provided, 
            use it instead of the configured one
            Parameters:
                lang_model : LanguageModel = Language Model to reevaluate
                db_manager : BasePersistency manager = (optional) Persistency manager object to retrieve data to use when evaluating
                eval_dataset_size : int = (optional) Size of the dataset to use when evaluating the model
                min_loss : float = (optional) Minimum acceptable loss, if the computed loss is greater than this treshold, returns true. Otherwise returns false
                should_retrain_fn : (float) -> bool = (optional) Function to check if, based on the loss of the model, it should be retrained. 
                                                    Should receive the loss and return a boolean telling if it should be retrained. 
                                                    Providing this argument will override min_loss argument
                fields : [str] = (optional) fields from scraped data instances to use during evaluation
            Return:
                If the given model should be retrained
        """
        # Sanity check
        assert eval_dataset_size > 0, "Eval dataset size should be a possitive number"
        assert min_loss >= 0, "min loss should be non negative"

        from c4v.classifier.language_model.language_model import LanguageModel

        # set up retrain function
        should_retrain_fn = should_retrain_fn or (lambda x: x > min_loss)

        # Set up db_manager
        db_manager = db_manager or self._persistency_manager

        # TODO should order by newer, but can't do it right now as we can't ensure
        # news dates can be parsed into datetime objects
        scraped_data = list(db_manager.get_all(limit=eval_dataset_size, scraped=True))
        if not scraped_data:
            raise ValueError(
                "No ScrapedData available for evaluation. You can get more data by using the `c4v scrawl` and `c4v scrape` functions, or using the microscope.Manager object"
            )

        # Set up dataset
        ds = LanguageModel.to_pt_dataset(
            lang_model.create_dataset_from_scraped_data(scraped_data, fields)
        )
        del scraped_data

        # Compute loss
        loss = lang_model.eval_accuracy(ds)
        return should_retrain_fn(loss)



from dynaconf import Dynaconf, Validator
import os

_HOME = os.environ.get("HOME")
settings = Dynaconf(
    envvar_prefix="C4V",
    settings_files=["./settings.toml", "config/.secrets.toml"],
    validators=[
        Validator("DATE_FORMAT", default="%Y-%m-%d %H:%M:%S.%f%z"),
        Validator("C4V_FOLDER", default=os.path.join(_HOME, ".c4v")),
        Validator(
            "LOCAL_SQLITE_DB_NAME", default="c4v_db.sqlite"
        ),  # Path to the local sqlite db file
        Validator("LOCAL_SQLITE_DB", default=os.path.join(_HOME, ".c4v/c4v_db.sqlite")),
        Validator(
            "DEFAULT_BASE_LANGUAGE_MODEL", default="BSC-TeMU/roberta-base-bne"
        ),  # Base language model for the classifier
        Validator(
            "DEFAULT_LANG_MODEL_MIN_LOSS", default=0.15
        ),  # Minimum acceptable loss for base language model, if the loss is greater, a new training is required
        Validator(
            "EXPERIMENTS_DIR", default=os.path.join(_HOME, ".c4v/experiments/")
        ),  # Default name for experiments folder
    ],
)

# `envvar_prefix` = export envvars with `export DYNACONF_FOO=bar`.
# `settings_files` = Load this files in the order.



"""
    This experiment is intended to train a classification 
    model with local data.
"""
# Python imports
import dataclasses
from typing import Dict, Any, List

# Local imports
from c4v.classifier.experiment import (
    ExperimentFSManager,
    BaseExperimentSummary,
    BaseExperimentArguments,
    BaseExperiment,
)
from c4v.classifier.classifier import Classifier
from c4v.scraper.scraped_data_classes.scraped_data import ScrapedData
from c4v.config import settings


@dataclasses.dataclass
class ClassifierArgs(BaseExperimentArguments):
    """
        Arguments passed to the model during training
    """

    training_args: Dict[str, Any] = dataclasses.field(
        default_factory=dict
    )  # training arguments passed to the trained object during training
    columns: List[str] = dataclasses.field(
        default_factory=lambda: ["text"]
    )  # fields from ScrapedData to use
    train_dataset_name: str = "classifier_training_dataset.csv"  # name of the training dataset stored in data.processed.huggingface
    confirmation_dataset_name: str = "classifier_confirmation_dataset.csv"  # name of the confirmation dataset stored in data.processed.huggingface
    description: str = None  # Optional description for this experiment
    val_dataset_proportion: float = 0.2  # How much in proportion for the training dataset take as eval dataset
    base_model_name: str = settings.default_base_language_model


@dataclasses.dataclass
class ClassifierSummary(BaseExperimentSummary):
    """
        Metrics returned by the classifier after every run 
    """

    eval_metrics: Dict[str, Any] = dataclasses.field(default_factory=dict)
    user_args: ClassifierArgs = ClassifierArgs()

    def __str__(self) -> str:
        super_str = super().__str__()
        # Add eval_metrics fields
        super_str += "EVAL METRICS:\n"
        super_str += (
            "\n".join(
                [
                    f"\t* {k} = {v}"
                    for (k, v) in self.eval_metrics["metrics_value"].items()
                ]
            )
            + "\n"
        )
        super_str += "USER ARGUMENTS:\n"
        super_str += "\tColumns: \n"
        super_str += (
            "\n".join([f"\t\t* {c}" for c in self.user_args.columns])
            if self.user_args.columns
            else "\t\t<No Columns Provided>"
        )
        super_str += "\n"
        super_str += f"\tTrain Dataset: {self.user_args.train_dataset_name}\n"
        super_str += (
            f"\tConfirmation Dataset: {self.user_args.confirmation_dataset_name}\n"
        )
        super_str += f"\tValidation Dataset Proportion: {self.user_args.val_dataset_proportion}\n"
        super_str += f"\tBase Model Name: {self.user_args.base_model_name}\n"
        super_str += "\tTraining Arguments:\n"
        super_str += (
            "\n".join(
                [f"\t\t* {k} = {v}" for (k, v) in self.user_args.training_args.items()]
            )
            + "\n"
        )

        return super_str


class ClassifierExperiment(BaseExperiment):
    """
        Use this experiment to run a classifier training. 
        You can check valid training arguments here:
        https://huggingface.co/transformers/main_classes/trainer.html#transformers.TrainingArguments
    """

    def __init__(
        self,
        experiment_fs_manager: ExperimentFSManager,
        classifier_instance: Classifier = None,
    ):
        # Initialize super class
        super().__init__(experiment_fs_manager=experiment_fs_manager)

        # Set up default values
        if not classifier_instance:
            classifier_instance = Classifier()

        # Set up classifier object
        classifier_instance.files_folder = (
            experiment_fs_manager.experiment_content_folder
        )

        self._classifier = classifier_instance

    @property
    def classifier(self) -> Classifier:
        """
            Configured classifier instance
        """
        return self._classifier

    def experiment_to_run(self, args: ClassifierArgs) -> ClassifierSummary:
        # Run a training process
        metrics = self._classifier.run_training(
            args.training_args,
            columns=args.columns,
            training_dataset=args.train_dataset_name,
            confirmation_dataset=args.confirmation_dataset_name,
            val_test_proportion=args.val_dataset_proportion,
            base_model_name=args.base_model_name,
        )
        summary = ClassifierSummary(
            eval_metrics=metrics, description=args.description, user_args=args
        )
        print(summary)
        return summary

    def classify(self, data: List[ScrapedData]) -> List[Dict[str, Any]]:
        """
            Classify this sentence using configured experiment
            Parameters:
                sentence : str = sentence or text to be classifier
            Return:
                A List of dicts with the resulting scraped data correctly labelled
                and its corresponding scores tensor for each possible label. Available fields:
                    + data : ScrapedData = resulting data instance after classification
                    + scores : torch.Tensor = Scores for each label returned by the classifier
        """
        return self._classifier.classify(data)

    def explain(
        self, sentence: str, html_file: str = None, additional_label: str = None
    ) -> Dict[str, Any]:
        """
            Explain given sentence using provided model
            Parameters:
                sentence : str = text to explain
                html_file : str = path to some html file to store human readable representation. If no provided, 
                                    it's ignored
                additional_label : str = Label to include in expalantion. If the predicted label is different 
                                         from this one, then explain how much this label was contributing to 
                                         its corresponding value. Ignored if not provided.
            Return:
                Dict with data for this explanation. For example:
                {   "scores" : 
                    [   
                        ('denuncian' , 0.98932),
                        ('falta'     , 0.78912),
                        ('de'        , 0.001231),
                        ('agua'      , 0.863781)
                    ],
                    "label" : "DENUNCIA_FALTA_DEL_SERVICIO",
        """
        return self._classifier.explain(sentence, html_file, additional_label)

    @classmethod
    def from_branch_and_experiment(
        cls,
        branch_name: str,
        experiment_name: str,
        classifier_instance: Classifier = None,
    ):
        """
            Create an experiment from a branch name, an experiment name, and a classifier instance
            Parameters:
                branch_name : str = Branch name for the experiment
                experiment_name : str = Experiment name
                classifier_instance : Classifier = optional classifier instance, will be defaulted if none was provided
        """
        fs_manager = ExperimentFSManager(branch_name, experiment_name)

        if classifier_instance:
            classifier_instance.files_folder_path = fs_manager.experiment_content_folder

        classifier_instance = classifier_instance or Classifier(
            fs_manager.experiment_content_folder
        )
        return cls(fs_manager, classifier_instance)



"""
    This class represents a model, you can extend it to have some common operations 
    for models, such as managing a files folder, or loading datasets from the data
    folder
"""
# Python imports
from pathlib import Path
from importlib import resources
import pandas as pd

# Third party imports
import torch
from transformers import PreTrainedTokenizer, AutoTokenizer, PreTrainedModel
from transformers.utils.dummy_pt_objects import AutoModel

# Local imports
from c4v.config import settings


class BaseModel:
    """
        Represents a model and contains useful utilities for models, such
        as managing files folders and loading datasets from the data folder.
    """

    RESULTS_EXPERIMENT_NAME = "results"
    LOGS_FOLDER_NAME = "logs"

    def __init__(
        self,
        files_folder_path: str = None,
        base_model_name: str = settings.default_base_language_model,
        use_cuda: bool = True,
    ) -> None:

        # Set up cuda
        self._device = (
            torch.device("cuda")
            if use_cuda and torch.cuda.is_available()
            else torch.device("cpu")
        )

        # Set up base model and files folder path
        self._base_model_name = base_model_name
        self.files_folder_path = files_folder_path

        # Init lazy tokenizer and model variables:
        self._model = None
        self._tokenizer = None

    @property
    def device(self):
        """
            Device used when evaluating the model
        """
        self._device

    @property
    def files_folder_path(self) -> str:
        """
            Folder to store files for a training process
        """
        return self._files_folder

    @files_folder_path.setter
    def files_folder_path(self, value: str):
        # Check that folder for internal files does exists
        if value and not Path(value).exists():
            raise ValueError(f"Given path does not exists: {value}")
        self._files_folder = value

    @property
    def logs_path(self) -> str:
        """
            Get path to logs for this experiment
            Return:
                path to log folder where we store logs for this experiment
        """
        if not self.files_folder_path:
            raise ValueError(
                "Could not create logs files, as no files folder is configured for this classifier object"
            )

        p = Path(self.files_folder_path, f"{self.LOGS_FOLDER_NAME}")

        # Create folder if does not exists
        if not p.exists():
            try:
                p.mkdir()
            except IOError as e:
                raise ValueError(
                    f"Could not create logs folder for path: {str(p)}, error: {e}"
                )

        return str(p)

    @property
    def results_path(self) -> str:
        """
            Get path to results for this experiment
            Return:
                path to results folder where we store results for this experiment
        """
        if not self.files_folder_path:
            raise ValueError(
                "Could not create results files, as no files folder is configured for this classifier object"
            )

        p = Path(self._files_folder, f"{self.RESULTS_EXPERIMENT_NAME}")

        # Create folder if does not exists
        if not p.exists():
            try:
                p.mkdir()
            except IOError as e:
                raise ValueError(
                    f"Could not create logs folder for path: {str(p)}, error: {e}"
                )

        return str(p)

    @property
    def tokenizer(self) -> PreTrainedTokenizer:
        """
            Internal tokenizer object. It's lazy-loaded, so it will
            be loaded once when it's called for the first time
        """
        if self._tokenizer == None:
            self._tokenizer = AutoTokenizer.from_pretrained(self._base_model_name)

        return self._tokenizer

    @property
    def model(self) -> PreTrainedModel:
        """
            Internal model object. It's lazy-loaded, so it will be loaded once when it's called for the first time
        """
        if self._model == None:
            self._model = AutoModel.from_pretrained(self._base_model_name)

        return self._model


class C4vDataFrameLoader:
    """
        Use this object to get data from the data folder as pandas dataframes
    """

    @staticmethod
    def get_from_raw(dataset_name: str) -> pd.DataFrame:
        """
            Get dataframe as a pandas dataframe, using a csv file stored in <project_root>/data/raw/huggingface
        """
        with resources.open_text("data.raw.huggingface", dataset_name) as f:
            return pd.read_csv(f)

    @staticmethod
    def get_from_processed(dataset_name: str) -> pd.DataFrame:
        """
            Get dataframe as a pandas dataframe, using a csv file stored in <project_root>/data/processed/huggingface
        """
        with resources.open_text("data.processed.huggingface", dataset_name) as f:
            return pd.read_csv(f)



import pandas as pd
import os
import sys
import csv
from datetime import datetime
import getopt
import argparse


class DataSampler:
    def __init__(
        self,
        tweets_dataset_path: str,
        sample_n: int,
        random_state: int,
        text_col: str,
        annotator_name: str,
        save_taken: bool,
        out_folder: str,
    ):
        """
        Creates a Data Sampler
        :param tweets_dataset_path: path of the tweets source
        :param sample_n: number of samples to extract
        :param random_state: random state
        :param text_col: name of the column with tweets
        :param annotator_name: name of the person that will be making annotations, this will also be the
            name of the folder where the data sample is stored.
        :param save_taken: a boolean that when True will store the ids of sampled tweets so future annotators
            do not extract already annotated ids
        :param out_folder: folder where all samples will be store. Each annotator should have a folder in this
            directory
        :return: None
        """

        self.tweets_dataset_path = os.path.relpath(tweets_dataset_path)

        # parent folder where .csv with tweets lives
        self.file_dir = os.path.dirname(tweets_dataset_path)
        # name of the .csv that contains all the tweets (output of sql in bigQuery)
        self.input_file = os.path.basename(tweets_dataset_path)
        # name of the .csv that contains the ids of annotated tweets
        self.tweets_annotated_ids = self.input_file.split(".")[0] + "-annotated_ids.csv"
        self.taken_ids = self.__setup_file_of_annotated_ids()

        self.save_taken = save_taken
        # self.brat_data_path = brat_data_path
        self.sample_n = sample_n
        self.random_state = random_state
        self.text_col = text_col
        self.annotator_name = annotator_name
        self.filename = None
        self.annotator_folder = os.path.dirname(
            f"{out_folder}{self.annotator_name}/"
        )  # output folder

        self.get_sample()

    def __setup_data_folder_for_annotator(self, basedir):
        """
        Creates the folder of the annotator, in case it does not exist.
        """
        if not os.path.exists(basedir):
            os.makedirs(basedir)
            print(
                f'\n\t* A folder for "{self.annotator_name}" has been created in path: {basedir}/ '
            )

    def __setup_file_of_annotated_ids(self) -> pd.DataFrame:
        """
        Reads the annotated ids from the csv.  In case the file does not exist.
        Returns an empty DataFrame, otherwise returns the contents of the file.
        """
        file_with_annotated_ids = os.path.join(self.file_dir, self.tweets_annotated_ids)

        if not os.path.isfile(file_with_annotated_ids):
            return pd.DataFrame({"id": []})

        return pd.read_csv(file_with_annotated_ids)

    def __save_annotated_ids(self, new_annotated_ids):

        if self.taken_ids.shape[0] > 0:
            # if there are ids, append the new to the list.
            annotated_ids = self.taken_ids.append(new_annotated_ids.to_frame())
        else:
            # if there are no existing ids, store this ones as new ones
            annotated_ids = new_annotated_ids

        annotated_ids.to_csv(
            f"{self.file_dir}/{self.tweets_annotated_ids}", index=False
        )
        print(f"File Updated: {self.file_dir}/{self.tweets_annotated_ids}")

    def __process_and_sample(self) -> pd.DataFrame:
        """
        Reads the csv with all the tweets, avoids getting a new sample with tweets that have
        already been annotated by saving a csv with all the tweet IDs that have already been tagged.
        The resulting csv will have for name the same of that original file with the suffix "-annotated_ids".

        :return: a dataframe that contains a sample of the tweets to be annotated.
        """

        original_tweets = pd.read_csv(self.tweets_dataset_path)

        # make sure the ids are of type integer and not interpreted as a float/scientific notation
        original_tweets["id"] = original_tweets["id"].astype(int)

        # drop the ids from the original_df, so when we sample we do not repeat an already tagged tweet
        available_tweets = original_tweets[
            ~original_tweets["id"].isin(self.taken_ids["id"].values)
        ]

        sampled_df = available_tweets.sample(
            random_state=self.random_state, n=self.sample_n
        )[["id", self.text_col]]

        # Reports the number of total tweets, the ones taken, the ones available and tha ones sampled
        print(f"Size of original file with tweets: {original_tweets.shape}")
        print(f"Size of the tweet IDs that have been taken: {self.taken_ids.shape}")
        print(f"Size of the available tweets: {available_tweets.shape}")
        print(f"Size of the sampled tweets {sampled_df.shape}")
        # print(f'sampled: {sampled_df}')

        if self.save_taken:
            self.__save_annotated_ids(sampled_df["id"])
        else:
            # prompts the user to mark the tweets as tagged
            user_response = str(
                input(
                    f"""
            Do you wish to set these sampled data as tagged?
            If you select <n> or any other key, you are at risk of extracting data that you have already worked on.
            Answer (y/n): """
                )
            )

            if user_response == "y":
                self.__save_annotated_ids(sampled_df["id"])
                print(
                    f"""
                File has been updated, and next time you require new data, we guarantee
                you will not have repeated data.
                """
                )
            else:
                print(
                    f"""
                You are at risk of extracting data that you have already worked on.  
                To avoid this, we recommend to tag the sampled data.
                """
                )

        return sampled_df

    def cleaner(self, df, is_pandas_series=False):
        """
        Helper function to do basic text cleaning operations.
        These include: Converting text to lower case, removing spanish accents,and removing links.
        -------------------------------------------------------------------------------------------
        PARAMS
            df: Dataframe or Pandas.Series object.
            text_col: String. Column to clean.
            is_pandas_series: Boolean, Optional. If df is pandas.Series

        """

        # to lower

        if not is_pandas_series:
            df[self.text_col] = df[self.text_col].str.lower()

            # Convert common spanish accents

            df[self.text_col] = df[self.text_col].str.replace("ú", "u")
            df[self.text_col] = df[self.text_col].str.replace("ù", "u")
            df[self.text_col] = df[self.text_col].str.replace("ü", "u")
            df[self.text_col] = df[self.text_col].str.replace("ó", "o")
            df[self.text_col] = df[self.text_col].str.replace("ò", "o")
            df[self.text_col] = df[self.text_col].str.replace("í", "i")
            df[self.text_col] = df[self.text_col].str.replace("ì", "i")
            df[self.text_col] = df[self.text_col].str.replace("é", "e")
            df[self.text_col] = df[self.text_col].str.replace("è", "e")
            df[self.text_col] = df[self.text_col].str.replace("á", "a")
            df[self.text_col] = df[self.text_col].str.replace("à", "a")
            df[self.text_col] = df[self.text_col].str.replace("ñ", "gn")

            # Remove Punctuation
            df[self.text_col] = df[self.text_col].str.replace("[\.\-:,\?]", " ")

            # Remove links
            df[self.text_col] = df[self.text_col].str.replace("http.+", " ")

            return df

        else:

            df = df.str.lower()

            # Convert common spanish accents

            df = df.str.replace("ú", "u")
            df = df.str.replace("ù", "u")
            df = df.str.replace("ü", "u")
            df = df.str.replace("ó", "o")
            df = df.str.replace("ò", "o")
            df = df.str.replace("í", "i")
            df = df.str.replace("ì", "i")
            df = df.str.replace("é", "e")
            df = df.str.replace("è", "e")
            df = df.str.replace("á", "a")
            df = df.str.replace("à", "a")
            df = df.str.replace("ñ", "gn")

            # Remove Punctuation
            df = df.str.replace("[\.\-:,\?]", " ")

            # Remove links
            df = df.str.replace("http.+", " ")

            return df

    def get_sample(self):

        original_df = self.__process_and_sample()

        original_df = self.cleaner(df=original_df)

        original_df = original_df[self.text_col].str.replace("\n", "")

        original_df = original_df.apply(
            lambda x: x.encode("ascii", "ignore").decode("ascii")
        )

        # Returns csv with annotator's name (e.g. Juanito perez) and with timestamp

        timestamp = datetime.now()
        formatted_timestamp = timestamp.strftime("%Y-%m-%d_%H%M%S")
        self.filename = f"{self.annotator_name}-sample_{self.sample_n}-randstate_{self.random_state}-{formatted_timestamp}.txt"

        try:
            # saves the tweets in txt
            original_df.to_csv(
                f"{self.annotator_folder}/{self.filename}",
                sep=" ",
                header=False,
                index=False,
                line_terminator="\n\n",
                quoting=csv.QUOTE_NONE,
                escapechar=" ",
            )

            # creates an empty .ann to use for annotations in brat
            open(
                f'{self.annotator_folder}/{self.filename.split(".")[0]}.ann', "a"
            ).close()

            print(
                f"""
    Success!

    Data was saved at {self.annotator_folder}/{self.filename}

    """
            )
            # !self.brat_data_path/bash ann_creator.sh
        except FileNotFoundError:
            print(
                f"""
    Error!

    There was an error with {self.annotator_folder}/{self.filename}.
    Please check that the folder in which we are saving the data does exist.
    Name of the folder: {self.annotator_folder.split('/')[-1]}

    """
            )


if __name__ == "__main__":
    """
    Example:

(1)
python data_sampler.py --path ../data_analysis/tagging-set-original_for_jupyter_tagging.csv --sample_size 30 --rand_state 19 --text_col full_text --annotator_name diego

   or 

(2)
python data_sampler.py -p=../data_analysis/tagging-set-original_for_jupyter_tagging.csv -s=30 -r=19 -tc=full_text --annotator_name diego

If case the annotator wishes not to repeat extracting tweets already annotated, a flag: -t ot --taken can be added
to the command (1) or (2) above.

For more information python data_sampler.py -h 

    """

    def annotator_folder_exists(arg_value):
        """
        See if the folder with the annotator's name has been created.
        :param arg_value: a string that represents the name of the annotator and at the same time the name
        of the folder in which the samples will be stored.
        :return: the string, if no errors are thrown
        """
        if not os.path.isdir(f"{OUTPUT_FOLDER}/{arg_value}"):
            msg = f'The folder of annotator "{arg_value}" has not been created. Please do.'
            raise argparse.ArgumentTypeError(msg)

        return arg_value

    def source_file_exists(arg_value):
        """
        See if the csv given as path exists.
        :param arg_value: a string that represents the filepath that contains all tweets. The source.
        :return: the filepath as string, if no errors are thrown
        """
        if not os.path.isfile(arg_value):
            msg = f'The provided path "{arg_value}" cannot be found. Please make sure it exists.'
            raise argparse.ArgumentTypeError(msg)

        return arg_value

    def check_input():
        parser = argparse.ArgumentParser(
            description="""
        Sample tweets dataframe for annotations in Brat.

        Example:
            python data_sampler.py --path ../data_analysis/tagging-set-original_for_jupyter_tagging.csv --sample_size 30 --rand_state 19 --text_col full_text --annotator_name diego
        """
        )
        parser.add_argument(
            "-p",
            "--path",
            type=source_file_exists,
            help="Path to csv file where tweets are located.",
        )

        parser.add_argument(
            "-t",
            "--taken",
            action="store_true",
            default=False,
            help="Determines whether to mark the sample of tweets obtained as tagged, so future "
            "annotators do not repeat annotating tweets that have already been taken.",
        )

        parser.add_argument(
            "-s", "--sample_size", type=int, help="Sample size to extract."
        )

        parser.add_argument(
            "-r", "--rand_state", type=int, help="Random State to used in the models."
        )

        parser.add_argument(
            "-tc", "--text_col", type=str, help="Column that contains rwe tweets text."
        )

        parser.add_argument(
            "--annotator_name",
            type=annotator_folder_exists,
            help="Name of the person that is annotating.  Also the name of the folder where "
            "to save the sampled data.",
        )

        return vars(parser.parse_args())

    # folder where all annotations will lay.
    OUTPUT_FOLDER = "../data/data_to_annotate/"

    args = check_input()
    # print(args)

    sample = DataSampler(
        tweets_dataset_path=args["path"],
        sample_n=args["sample_size"],
        random_state=args["rand_state"],
        text_col=args["text_col"],
        annotator_name=args["annotator_name"],
        save_taken=args["taken"],
        out_folder=OUTPUT_FOLDER,
    )



"""
    The classifier requieres a base language model that might need 
    to be pre trained for Venezuelan Spanish. With this class you can
    check how good the base model is for the task and pre train it if 
    necessary
"""
# Third party imports
from torch.utils.data import Dataset
from transformers import (
    EvalPrediction,
    AutoTokenizer,
    AutoModelForMaskedLM,
    TrainingArguments,
    Trainer,
    BatchEncoding,
    PreTrainedModel,
)
import pandas as pd
import torch

# Python imports
from typing import Callable, Iterable, List, Any, Dict
import dataclasses as dc
import tempfile

# Local imports
from c4v.scraper.scraped_data_classes.scraped_data import ScrapedData
from c4v.classifier.base_model import BaseModel


class LanguageModel(BaseModel):
    """
        This is the base model that gets trained for the classifier downstream 
        task. Use this class to evaluate accuracy for a base language model, and 
        retrain its embeddings layer if needed
    """

    def create_dataset_from_scraped_data(
        self,
        data: Iterable[ScrapedData],
        fields: List[str] = ["content"],
        tokenizer: Any = None,
    ) -> BatchEncoding:
        """
            Creates a dataset from ScrapedData instances, ready to be used for training or 
            evaluation, containing both the masked version and its corresponding actual answer.

            Parameters:
                data : Iterable[ScrapedData] = Data to use in the dataset 
                fields : [str] = (optional) names of fields in each instance to use when extracting text
                tokenizer : Any = (optional) A huggingface tokenizer used to tokenize text into input ids, use the 
                                  configured model's tokenizer by default, provide this field to override it
            Return:
                A dataset with masked and correct answer based on the given data as rows
        """

        # Sanity check: provided fields are valid fields
        valid_fields = set(f.name for f in dc.fields(ScrapedData))
        fields_set = set(fields)

        # Use only fields that are both in the provided list and the valid fields list
        fields_to_use = valid_fields.intersection(fields_set)

        # Raise error if can't use any field
        if not fields_to_use:
            raise ValueError(
                f"No valid field provided in 'fields' field of 'LanguageModel.create_dataset_from_scraped_data' function, provided list: {fields}. Valid fields: {list(valid_fields)}"
            )

        # Extract text to be used
        text = [
            "\n".join([str(d.__getattribute__(attr)) for attr in fields_to_use])
            for d in data
        ]

        # Clean text
        text = [
            "\n".join(
                [s for s in [s.strip() for s in d.split("\n")] if s != ""]
            ).replace("\n", ". ")
            for d in text
        ]

        # Set up tokenizer:
        tokenizer = tokenizer or AutoTokenizer.from_pretrained(self._base_model_name)

        # Tokenize text
        tokenized_text = tokenizer(
            text,
            max_length=512,
            padding="max_length",
            truncation=True,
            return_tensors="pt",
        )  # TODO deberíamos configurar esto con argumentos en la funcion

        # Valid answers & attention mask
        labels = torch.tensor(tokenized_text["input_ids"])

        # Input ids: fed to the model as the masked version
        input_ids = labels.detach().clone()

        # generate matrix of the same size of random numbers, we're going to use it
        # to mask some words for the model
        rand = torch.rand(input_ids.shape)

        # Create mask array
        masked_tokens_array = rand < 0.15

        for (_, tk_value) in tokenizer.special_tokens_map.items():

            masked_tokens_array = masked_tokens_array * (
                input_ids != tokenizer.vocab[tk_value]
            )

        # Convert selected tokens to mask
        for i in range(input_ids.shape[0]):
            # get indices of mask positions from mask array
            selection = torch.flatten(masked_tokens_array[i].nonzero()).tolist()
            input_ids[i, selection] = tokenizer.vocab[
                tokenizer.mask_token
            ]  # Token <mask> id

        tokenized_text["labels"] = labels
        tokenized_text["input_ids"] = input_ids

        return (
            tokenized_text  # type = transformers.tokenization_utils_base.BatchEncoding
        )

    @staticmethod
    def to_pt_dataset(batch: BatchEncoding) -> Dataset:
        """
            Turn batch enconding (as the one created by create_dataset_from_scraped_data) into a pytorch dataset
        """

        class _Dataset(Dataset):
            def __init__(self, encodings):
                self.encodings = encodings

            def __getitem__(self, idx):
                return {
                    key: torch.tensor(val[idx]) for key, val in self.encodings.items()
                }

            def __len__(self):
                return len(self.encodings.input_ids)

        return _Dataset(batch)

    def eval_accuracy(
        self, dataset: Dataset, model: Any = None, batch_size: int = 1
    ) -> float:
        """
            Try to eval accuracy of this language model for the given dataset.
            Parameters:
                dataset : BatchEnconding = dataset with data to use for evaluation, should contain the 
                                            input ids vector to unmask, with masked words. Shoud provide the attention_mask,
                                            inputs_ids, and labels fields
                model : Any = Huggingface model for masked language modeling, be default we use the model refered
                              by the configured model name, you can override it by providing this field
                batch_size : int = batch size to send to gpu for evaluation
            Return:
                The loss for the given dataset
        """
        # set up model
        model = model or self.model

        # Load model
        model.to(self._device)

        with tempfile.TemporaryDirectory() as temp_dir:
            args = TrainingArguments(
                per_device_eval_batch_size=batch_size,
                output_dir=temp_dir,
                per_device_train_batch_size=1,
                eval_accumulation_steps=1,
                do_train=False,
                prediction_loss_only=True,
            )
            trainer = Trainer(args=args, model=model, eval_dataset=dataset)
            # Evaluate model
            outputs = trainer.evaluate()

        return outputs["eval_loss"]

    @property
    def _default_train_args(self) -> Dict[str, Any]:
        """
            A default version of training arguments as a dict
        """

        default = {
            "output_dir": self.results_path,
            "num_train_epochs": 1,
            "per_device_train_batch_size": 10,
            "per_device_eval_batch_size": 10,
            "warmup_steps": 10000,
            "weight_decay": 0.01,
            "logging_dir": self.logs_path,
            "save_total_limit": 1,
        }

        return default

    @property
    def model(self) -> PreTrainedModel:
        """
            Internal model object. It's lazy-loaded, so it will be loaded once when it's called for the first time
        """
        if self._model == None:
            self._model = AutoModelForMaskedLM.from_pretrained(
                self._base_model_name
            )  # using specific desired model

        return self._model

    def _override_train_args(self, new_args: Dict[str, Any] = {}) -> Dict[str, Any]:
        """
            Return the default args dict, with overriden settings specified as the ones specified in 
            input dict
        """

        default = self._default_train_args
        for (k, v) in new_args.items():
            default[k] = v

        return default

    def train_and_save_model(
        self,
        train_dataset: Dataset,
        eval_dataset: Dataset,
        train_args: Dict[str, Any],
        output_dir_path: str = None,
        logging_dir_path: str = None,
        path_to_save_checkpoint: str = None,
        compute_metrics: Callable[[EvalPrediction], Dict[str, Any]] = None,
        model: Any = None,
    ) -> Trainer:
        """
            Train a given model, or the one for the provided model 
            Parameters: 
                dataset : Dataset = training data with input_ids and labels
                model : Any = model to train, provide this field to override the configured model
            Return:
                Trainer object with the fine tuned model 
        """

        if output_dir_path:
            train_args["output_dir"] = output_dir_path
        if logging_dir_path:
            train_args["logging_dir"] = logging_dir_path

        args = TrainingArguments(**self._override_train_args(train_args))

        trainer = Trainer(
            model=model,
            args=args,
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            compute_metrics=compute_metrics,
        )

        trainer.train()

        trainer.model.save_pretrained(path_to_save_checkpoint or self.files_folder_path)

        return trainer

    def evaluate_metrics(self, trainer: Trainer, val_dataset: Dataset) -> pd.DataFrame:
        """
            Compute metrics as a dataframe for this trainer object using the validation dataset
            Parameters:
                trainer : Trainer = trainer object already trained and ready to test
                val_dataset : Dataset = data used for validation
            Return:
                A dataframe with output of a validation
        """
        ## Evaluate Metrics
        metrics = trainer.evaluate(val_dataset)
        metrics_df = pd.DataFrame.from_dict(
            metrics, orient="index", columns=["metrics_value"]
        )
        return metrics_df

    def run_training(
        self,
        train_dataset: Dataset,
        eval_dataset: Dataset,
        confirmation_dataset: Dataset,
        train_args: Dict[str, Any] = None,
        model_name: str = None,
    ) -> pd.DataFrame:
        """
            Run an experiment specified by given train_args, and write a summary if requested so
            Parameters:
                train_dataset : Dataset = dataset to use during training
                eval_dataset : Dataset = dataset to use during training, when evaluating a model
                confirmation_dataset : Dataset = dataset to use after training, when validating a model
                train_args : Dict[str, Any] = (optional) arguments passed to trainig arguments
                model_name : str = (optional) Base model name for selecting next model to use
            Return:
                Classifier metrics as a Dataframe
        """

        # check that you have a folder where to store results
        if not self.files_folder_path:
            raise ValueError(
                "Can't train in a Classifier without a folder for local data"
            )

        # Prepare dataframe and load model
        assert model_name or self._base_model_name, "Model to train not provided"
        model = (
            AutoModelForMaskedLM.from_pretrained(model_name)
            if model_name
            else self.model
        )

        # Fine tune the model
        fine_tuned_model_trainer = self.train_and_save_model(
            model=model,
            output_dir_path=self.results_path,
            logging_dir_path=self.logs_path,
            path_to_save_checkpoint=self.files_folder_path,
            train_args=self._override_train_args(train_args or {}),
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
        )

        # Save tokenizer too
        self.tokenizer.save_pretrained(self.files_folder_path)

        # Get the metrics from the model
        metrics_df = self.evaluate_metrics(
            trainer=fine_tuned_model_trainer, val_dataset=confirmation_dataset
        )

        return metrics_df



from c4v.classifier.classifier import Classifier
from c4v.classifier.classifier_experiment import ClassifierExperiment, ClassifierArgs, ClassifierSummary
from c4v.classifier.experiment import BaseExperiment, BaseExperimentArguments, BaseExperimentSummary


from c4v.scraper.crawler.crawlers.base_crawler import BaseCrawler
from c4v.scraper.crawler.crawlers.primicia_crawler import PrimiciaCrawler
from c4v.scraper.crawler.crawlers.el_pitazo_crawler import ElPitazoCrawler


# Import interface class
from typing import Dict, Iterator, List
from c4v.scraper.persistency_manager import BasePersistencyManager
from c4v.scraper import ScrapedData
from dataclasses import fields
import sys

class DictManager(BasePersistencyManager):
    """
        Persistency Manager class that stores data in a dict object
    """

    def __init__(self):
        super().__init__() # Don't forget to call the parent class constructor
        # We will store ScrapedData instances in a dict, using its 
        # url as a key
        self._stored_data : Dict[str, ScrapedData] = {}

    def get_all(self, limit: int = -1, scraped: bool = None, order_by: List[str] = None) -> Iterator[ScrapedData]:        

        # Remember to do some sanity check
        valid_fields = {f.name for f in fields(ScrapedData)}

        order_by = order_by or [] 
        for order in order_by:
            if not (order and order[0] in ["-", "+"] and order[1:] in  valid_fields):
                raise ValueError("not valid order provided: " + order) 
        
        # Get actual data, filtering by scraped or not
        # This lambda is checking if the data instance should be included in the query.
        # We assume that an instance is not scraped when its last_scraped field is not provided
        # (AKA it wasn't scraped in any moment)
        goes_in = lambda x: scraped == None or \
                            (x.last_scraped and scraped) or \
                            (not x.last_scraped and not scraped) 
        data = [d for d in self._stored_data.values() if goes_in(d)]

        # Now sort it as requested    
        for field in reversed(order_by): # order in inverse key order so you preserve multi sorting
            asc = field[0] == "+" 
            data.sort(key=lambda d: d.__getattribute__(field[1:]), reverse=not asc)
        
        # Set up limit 
        # All elements by default
        n_elems = len(data)
        limit = min(limit, n_elems) if limit > 0 else n_elems

        for i in range(limit):
            yield data[i]

    def filter_known_urls(self, urls: List[str]) -> List[str]:
        # Just return the ones that are not stored in our dict
        return [url for url in urls if not self._stored_data.get(url)]

    def filter_scraped_urls(self, urls: List[str]) -> List[str]:
        # Just return the ones that are either not stored, or stored but not yet scraped
        return [
            url for url in urls 
                if not self._stored_data.get(url) or\
                not self._stored_data[url].last_scraped
            ]

    def was_scraped(self, url: str) -> bool:
        # Return true if it's stored in DB and it was scraped at some point
        return bool(self._stored_data.get(url) and self._stored_data[url].last_scraped)

    def save(self, url_data: List[ScrapedData]):
        # Add data instance to dict, use its url as key
        for d in url_data:
            self._stored_data[d.url] = d

    def delete(self, urls: List[str]):
        # Just remove these urls from storage dict
        for url in urls:
            if self._stored_data.get(url):
                del self._stored_data[url]


"""
    Scraper to get data from Primicia
"""
# Internal imports
from c4v.scraper.scrapers.base_scrapy_scraper import BaseScrapyScraper
from c4v.scraper.spiders.primicia import PrimiciaSpider


class PrimiciaScraper(BaseScrapyScraper):
    """
        Scrapes data from ElPitazo, relies in 
        scrapy for this.
    """

    intended_domain = "primicia.com.ve"
    spider = PrimiciaSpider



import os, sys
from datetime import datetime
import pandas as pd
import numpy as np

from sklearn.metrics import f1_score
from sklearn.metrics import accuracy_score
from sklearn.metrics import hamming_loss
from sklearn.metrics import precision_score
from sklearn.metrics import recall_score

from skmultilearn.problem_transform import BinaryRelevance
from skmultilearn.problem_transform import ClassifierChain
from skmultilearn.problem_transform import LabelPowerset
from sklearn.naive_bayes import BernoulliNB
from sklearn.linear_model import LogisticRegression, SGDClassifier
from sklearn.svm import LinearSVC
from skmultilearn.adapt import MLkNN
from scipy.sparse import lil_matrix

from c4v.data.data_loader import BratDataLoader


class ModelWrapper:
    def __init__(self, model, x_train, y_train, name, type="normal"):
        self.model_class_name = model.__class__
        self.model_name = name
        self.model_x = model
        self.type_ = type

        if self.type_ == "cc":
            self.selected_labels = y_train.columns[
                y_train.sum(axis=0, skipna=True) > 0
            ].tolist()
            y_train = y_train.filter(self.selected_labels, axis=1)
        if self.type_ in ["knn", "logistic"]:
            x_train, y_train = (
                lil_matrix(x_train).toarray(),
                lil_matrix(y_train).toarray(),
            )

        self.model_x.fit(x_train, y_train)

    def get_predictions(self, x_test1):

        if self.type_ in ["knn", "logistic"]:
            x_test = lil_matrix(x_test1).toarray()
        else:
            x_test = x_test1

        return self.model_x.predict(x_test)

    def get_metrics(self, x_test1, y_test1) -> dict:
        if self.type_ == "cc":
            y_test = y_test1.filter(self.selected_labels, axis=1)
        else:
            y_test = y_test1

        y_pred = self.get_predictions(x_test1)

        return {
            "accuracy": accuracy_score(y_test, y_pred),
            "hamming_loss": hamming_loss(y_test, y_pred),
            "f1_score_micro": f1_score(y_test, y_pred, average="micro"),
            "f1_score_macro": f1_score(y_test, y_pred, average="macro"),
            "precision_micro": precision_score(y_test, y_pred, average="micro"),
            "precision_macro": precision_score(y_test, y_pred, average="macro"),
            "recall_micro": recall_score(y_test, y_pred, average="micro"),
            "recall_macro": recall_score(y_test, y_pred, average="macro"),
        }

    def report_as_dict(self, x_test, y_test):
        metrics = self.get_metrics(x_test, y_test)
        # print('getting metris of ', metrics)
        metrics["model_name"] = self.model_name
        metrics["model_class_name"] = self.model_x.__class__.__name__
        return metrics


class ModelAnalyzer:
    def __init__(self, file_names: list):
        self.data: BratDataLoader
        self.performance: pd.DataFrame
        self.analysis: pd.DataFrame

        self.generate_report(file_names)

    def __analysis_tweets_classification(
        self, model_x_list: list, data: BratDataLoader
    ) -> pd.DataFrame:
        analysis_report = list()
        for model in model_x_list:
            expected = data.y_test
            predicted = model.get_predictions(data.X_test)
            all_tw, train_tw, test_tw = data.get_X_as_text()
            test_tw_df = test_tw.to_frame()
            predicted_df = pd.DataFrame(
                predicted.toarray(),
                columns=["pred_" + name for name in expected.columns.to_list()],
            )

            assert_df = pd.DataFrame(
                predicted.toarray() == expected.to_numpy(),
                columns=["assert_" + name for name in expected.columns.to_list()],
            )
            assert_df = assert_df.mask(assert_df == True, 1).mask(assert_df == False, 0)

            model_wrapper_df = pd.DataFrame(
                [model.model_class_name for _ in range(data.X_test.shape[0])],
                columns=["model_class_name"],
            )
            model_name_df = pd.DataFrame(
                [model.model_name for _ in range(data.X_test.shape[0])],
                columns=["model_name"],
            )

            end = pd.concat(
                [
                    test_tw_df.reset_index(drop=True),
                    expected.reset_index(drop=True),
                    predicted_df,
                    assert_df,
                    model_name_df,
                    model_wrapper_df,
                ],
                axis=1,
            )

            analysis_report.append(end)

        return pd.concat(analysis_report)

    def generate_report(self, file_names: list) -> None:
        """
        Generates pandas data frame objects with performance report using a different model. And also generates a
        table that from which we will be able to recognize what type of data our models do not correctly predict.
        :param file_names: list of file paths that contain the annotated data and the test data (tweets)
        """
        # loading the data
        random_state = 42
        self.data = data = BratDataLoader(file_names, binary=True,)
        print(f"the shape of my data.X: {data.X.shape}")
        data.preprocess(random_state=random_state)

        # Creating the models
        br_BernoulliNB_classifier = BinaryRelevance(BernoulliNB())
        lp_BernoulliNB_classifier = LabelPowerset(BernoulliNB())
        cc_BernoulliNB_classifier = ClassifierChain(BernoulliNB())
        br_LogisticR_classifier = BinaryRelevance(
            LogisticRegression(random_state=random_state)
        )
        lp_LogisticR_classifier = LabelPowerset(
            LogisticRegression(random_state=random_state)
        )
        cc_LogisticR_classifier = ClassifierChain(
            LogisticRegression(random_state=random_state)
        )
        br_SVM_classifier = BinaryRelevance(LinearSVC(random_state=random_state))
        lp_SVM_classifier = LabelPowerset(LinearSVC(random_state=random_state))
        cc_SVM_classifier = ClassifierChain(LinearSVC(random_state=random_state))
        br_GradDesc_classifier = BinaryRelevance(
            SGDClassifier(random_state=random_state, loss="log", penalty="elasticnet")
        )
        lp_GradDesc_classifier = LabelPowerset(
            SGDClassifier(random_state=random_state, loss="log", penalty="elasticnet")
        )
        cc_GradDesc_classifier = ClassifierChain(
            SGDClassifier(random_state=random_state, loss="log", penalty="elasticnet")
        )
        ml_classifier = MLkNN(k=4)

        # Creating a list of model wrappers. Training happens after each wrapper is instantiated.
        models = [
            ModelWrapper(
                ml_classifier, data.X_train, data.y_train, name="Multi kNN", type="knn"
            ),
            ModelWrapper(
                br_BernoulliNB_classifier,
                data.X_train,
                data.y_train,
                name="BR_Bayes BernoulliNB",
            ),
            ModelWrapper(
                br_LogisticR_classifier, data.X_train, data.y_train, name="BR_LogisticR"
            ),
            ModelWrapper(
                br_SVM_classifier, data.X_train, data.y_train, name="BR_LinearSVC"
            ),
            ModelWrapper(
                br_GradDesc_classifier,
                data.X_train,
                data.y_train,
                name="BR_GradDesc SGDCClassifier",
            ),
            ModelWrapper(
                lp_BernoulliNB_classifier,
                data.X_train,
                data.y_train,
                name="LP_Bayes BernoulliNB",
            ),
            ModelWrapper(
                lp_LogisticR_classifier, data.X_train, data.y_train, name="LP_LogisticR"
            ),
            ModelWrapper(
                lp_SVM_classifier, data.X_train, data.y_train, name="LP_LinearSVC"
            ),
            ModelWrapper(
                lp_GradDesc_classifier,
                data.X_train,
                data.y_train,
                name="LP_GradDesc SGDCClassifier",
            ),
            ModelWrapper(
                cc_BernoulliNB_classifier,
                data.X_train,
                data.y_train,
                name="CC_Bayes BernoulliNB",
                type="cc",
            ),
            ModelWrapper(
                cc_LogisticR_classifier,
                data.X_train,
                data.y_train,
                name="CC_LogisticR",
                type="cc",
            ),
            ModelWrapper(
                cc_SVM_classifier,
                data.X_train,
                data.y_train,
                name="CC_LinearSVC",
                type="cc",
            ),
            ModelWrapper(
                cc_GradDesc_classifier,
                data.X_train,
                data.y_train,
                name="CC_GradDesc SGDCClassifier",
                type="cc",
            ),
        ]
        # getting table with all results, ready to use for micro analysis
        self.analysis = self.__analysis_tweets_classification(models, data)

        # getting report of models performance
        self.performance = pd.DataFrame(
            [m.report_as_dict(data.X_test, data.y_test) for m in models]
        )

    def save_reports(self):
        # TODO: Ver como hacer el logging en data science antes de implementar esto!
        # Storing on disk the csv corresponding to the report, a timestamp is added as part of the file's name.
        # the name of the original annotation file(s) will be saved on the csv with "-source" as suffix.
        self.__save_report_on_disk(self.data.source_as_list(), self.data.X.shape[0])

    def __save_report_on_disk(self, file_names: list, data_set_size: int) -> None:
        out_filename = (
            f'report-{datetime.now().strftime("%Y_%m_%d_%H%M%S")}-{data_set_size}t'
        )
        reports_folder = "reports/ml_models_performance_analysis"
        self.performance.sort_values(by="accuracy").to_csv(
            os.path.join(reports_folder, out_filename + "-performance.csv")
        )

        pd.DataFrame({"source": file_names}).to_csv(
            os.path.join(reports_folder, out_filename + "-sources.csv")
        )

        self.analysis.to_csv(
            os.path.join(reports_folder, out_filename + "-analysis.csv")
        )


def unannotated_tweets(with_annotation: str) -> None:
    """
    Shows tweets that have not been captured by the load_data class. This function makes use of BratDataLoader
    :param with_annotation: path from current location to the annotated pair .txt and .ann
     no extension required.
    :return: None
    """
    with open(f"{with_annotation}.txt") as f:
        lines = f.read().replace("\n\n", "\n").split("\n")[:-1]

    # Careful here,
    # if .X were to be used after .preprocess(), it might have been converted into a vector, no longer text.
    annotated_data = BratDataLoader([with_annotation]).X
    o_lines = [line.replace("\n", "") for line in annotated_data]

    missed_values = [line for line in lines if line not in o_lines]

    print(
        f"**********************\nFile: {with_annotation}\nMissing: {len(missed_values)}\n"
    )
    for m in missed_values:
        print("-\t", m, "\n")



# External imports
import scrapy
from scrapy.http import Response

# Local imports
from .. import utils 
from c4v.scraper.scraped_data_classes.primicia_scraped_data import PrimiciaData


class PrimiciaSpider(scrapy.Spider):
    """
    Spider to scrape Primicia data
    """

    name = "primicia"

    start_urls = []

    # Selectors of Primicia
    TITLE_SELECTOR = "h1.title-single ::text"
    BODY_SELECTOR = ".credit-photoleyends + div ::text"
    DATE_SELECTOR = "#namesingle + div small ::text"
    AUTHOR_SELECTOR = "#namesingle a.author-link ::text"
    TAGS_SELECTOR = ".badge.badge-pill.badge-primicia ::text"
    CATEGORIES_SELECTOR = None

    def parse(self, response: Response, **kwargs) -> PrimiciaData:
        """
        Return a PrimiciaData instance, describing data possible scraped for this page
        Parameters:
            + response : Response = Scrapy response object
        Return:
            Data scraped for this page
        """

        title = response.css(self.TITLE_SELECTOR).get() or ""
        body = self._get_body(response, self.BODY_SELECTOR)
        date = response.css(self.DATE_SELECTOR).get() or ""
        author = response.css(self.AUTHOR_SELECTOR).get() or ""
        tags = response.css(self.TAGS_SELECTOR).getall() if self.TAGS_SELECTOR else []
        categories = (
            response.css(self.CATEGORIES_SELECTOR).getall()
            if self.CATEGORIES_SELECTOR
            else []
        )

        return PrimiciaData(
            tags=tags,
            categories=categories,
            title=title,
            author=author,
            date=date,
            body=body,
            url=response.url,
            last_scraped=utils.get_datetime_now(),
        )

    def _get_body(self, response: Response, selector) -> str:
        """
        Get article body as a single unicode string
        """
        body = response.css(selector).getall()
        body = "\n".join(body)

        return body.strip()



"""
    This experiment will run a base language model training
"""
# Python imports
import dataclasses
from typing import Any, Dict, List

# Third party imports
from torch.utils.data import Dataset
import pandas as pd

# Local imports
from c4v.classifier.experiment import (
    BaseExperiment,
    BaseExperimentSummary,
    BaseExperimentArguments,
    ExperimentFSManager,
)
from c4v.classifier.language_model.language_model import LanguageModel


@dataclasses.dataclass
class LanguageModelExperimentArguments(BaseExperimentArguments):
    """
        Arguments for Language model experiment
        You can check valid training arguments here:
        https://huggingface.co/transformers/main_classes/trainer.html#transformers.TrainingArguments
        Note:
            You can easlily build a dataset using the LanguageModel.create_dataset_from_scraped_data function 
                 
    """

    model_name: str = None  # Base model name to load
    use_cuda: bool = True  # If should use cuda when available. Ignored when not available
    train_dataset: Dataset = None  # Training dataset
    eval_dataset: Dataset = None  # Evaluation dataset (used during training)
    confirmation_dataset: Dataset = None  # Confirmation dataset (not used during training, used afterwards to validate training)
    train_args: Dict[str, Any] = dataclasses.field(
        default_factory=dict
    )  # Training arguments passed to the Trainer object
    description: str = None  # Optional description


@dataclasses.dataclass
class LanguageModelExperimentSummary(BaseExperimentSummary):
    """
        Language model experiment summary
    """

    user_args: LanguageModelExperimentArguments = LanguageModelExperimentArguments()
    result: pd.DataFrame = None

    def __str__(self) -> str:
        s = super().__str__().rstrip() + "\n"

        # User arguments
        s += "USER ARGS:\n"
        s += f"\tmodel_name : {self.user_args.model_name}\n"
        s += f"\tuse_cuda : {self.user_args.use_cuda}\n"
        s += "\ttrain_args : {\n"
        for (k, v) in self.user_args.train_args.items():
            s += f"\t\t{k} : {v}\n"
        s += "\t}\n"

        # Result:
        s += "RESULTS:\n"
        s += str(self.result)

        return s


class LanguageModelExperiment(BaseExperiment):
    """
        Experiment to run a training for the language model
    """

    def experiment_to_run(
        self, args: LanguageModelExperimentArguments
    ) -> LanguageModelExperimentSummary:
        lang_model = LanguageModel(
            files_folder_path=self._experiment_fs_manager.experiment_content_folder,
            base_model_name=args.model_name,
            use_cuda=args.use_cuda,
        )

        metrics_df = lang_model.run_training(
            train_dataset=args.train_dataset,
            eval_dataset=args.eval_dataset,
            train_args=args.train_args,
            confirmation_dataset=args.confirmation_dataset,
        )
        summary = LanguageModelExperimentSummary(
            description=args.description, result=metrics_df, user_args=args
        )
        return summary
